import {log} from "@blitzjs/display"
import {Stage, transform} from "@blitzjs/file-pipeline"
import {OverrideTriage} from "@blitzjs/file-pipeline/src/helpers/work-optimizer"
import * as fs from "fs-extra"
import {debounce, partition} from "lodash"
import {join} from "path"
import File from "vinyl"

function makeDebouncedWriter(path: string, ms = 100): (contents: string) => void {
  return debounce((contents: string) => {
    fs.outputFileSync(path, contents, {encoding: "utf-8"})
  }, ms)
}

type Parameter = {
  name: string
  optional: boolean
}
interface Route {
  name: string
  parameters: Parameter[]
  multipleParameters: Parameter[]
}

export function parseDefaultExportName(contents: string): string | null {
  const result = contents.match(/export\s+default(?:\s+(?:const|let|class|var|function))?\s+(\w+)/)
  if (!result) {
    return null
  }

  return result[1] ?? null
}

const isPage = (path: string) =>
  path.includes(join("pages", "/")) &&
  !path.includes(join("pages", "api", "/")) &&
  !path.includes(join("/", "_app.")) &&
  !path.includes(join("/", "_document."))

function dedupeBy<T, K>(arr: T[], by: (v: T) => K): T[] {
  const allKeys = arr.map(by)
  return arr.filter((v, index) => {
    const key = by(v)
    const first = allKeys.indexOf(key)
    const last = allKeys.lastIndexOf(key)

    if (first !== last && first !== index) {
      const {0: firstPath} = arr[first] as any
      const {0: lastPath} = arr[last] as any
      throw Error(
        `The page component is named "${key}" on both the ${firstPath} and ${lastPath} routes. The page component must have a unique name across all routes, so change the component name on one of those routes to avoid conflict.`,
      )
    }

    return true
  })
}

export function generateManifest(
  routes: Record<string, Route>,
): {implementation: string; declaration: string} {
  const routesWithoutDuplicates = dedupeBy(Object.entries(routes), ([_path, {name}]) => name)

  const implementationLines = routesWithoutDuplicates.map(
    ([path, {name}]) => `${name}: (query) => ({ pathname: "${path}", query })`,
  )

  const declarationLines = routesWithoutDuplicates.map(
    ([_path, {name, parameters, multipleParameters}]) => {
      if (parameters.length === 0 && multipleParameters.length === 0) {
        return `${name}(query?: ParsedUrlQueryInput): RouteUrlObject`
      }

      return `${name}(query: { ${[
        ...parameters.map(
          (param) => param.name + (param.optional ? "?" : "") + ": string | number",
        ),
        ...multipleParameters.map(
          (param) => param.name + (param.optional ? "?" : "") + ": (string | number)[]",
        ),
      ].join("; ")} } & ParsedUrlQueryInput): RouteUrlObject`
    },
  )

  return {
    implementation:
      "exports.Routes = {\n" + implementationLines.map((line) => "  " + line).join(",\n") + "\n}",
    declaration: `
import type { ParsedUrlQueryInput } from "querystring"
import type { RouteUrlObject } from "blitz"

export const Routes: {
${declarationLines.map((line) => "  " + line).join(";\n")};
}`.trim(),
  }
}

function removeSquareBracketsFromSegments(value: string): string

function removeSquareBracketsFromSegments(value: string[]): string[]

function removeSquareBracketsFromSegments(value: string | string[]): string | string[] {
  if (typeof value === "string") {
    return value.replace("[", "").replace("]", "")
  }
  return value.map((val) => val.replace("[", "").replace("]", ""))
}

const squareBracketsRegex = /\[\[.*?\]\]|\[.*?\]/g

export function parseParametersFromRoute(
  path: string,
): Pick<Route, "parameters" | "multipleParameters"> {
  const parameteredSegments = path.match(squareBracketsRegex) ?? []
  const withoutBrackets = removeSquareBracketsFromSegments(parameteredSegments)

  const [multipleParameters, parameters] = partition(withoutBrackets, (p) => p.includes("..."))

  return {
    parameters: parameters.map((value) => {
      const containsSquareBrackets = squareBracketsRegex.test(value)
      if (containsSquareBrackets) {
        return {
          name: removeSquareBracketsFromSegments(value),
          optional: true,
        }
      }

      return {
        name: value,
        optional: false,
      }
    }),
    multipleParameters: multipleParameters.map((param) => {
      const withoutEllipsis = param.replace("...", "")
      const containsSquareBrackets = squareBracketsRegex.test(withoutEllipsis)

      if (containsSquareBrackets) {
        return {
          name: removeSquareBracketsFromSegments(withoutEllipsis),
          optional: true,
        }
      }

      return {
        name: withoutEllipsis,
        optional: false,
      }
    }),
  }
}

/**
 * Will resolve the real node_modules root.
 * We're not fooled by you, `yarn workspace`!
 */
function findNodeModulesRoot(src: string) {
  let nodeModules = join(src, "node_modules")
  let includesBlitzPackage = fs.existsSync(join(nodeModules, "blitz"))
  let count = 0
  while (!includesBlitzPackage) {
    // Check for node_modules at the next level up
    nodeModules = join(nodeModules, "../../node_modules")
    includesBlitzPackage = fs.existsSync(join(nodeModules, "blitz"))
    count++
    if (count > 5) {
      log.warning(
        "We couldn't determine your actual node_modules location, so defaulting to normal location",
      )
      nodeModules = join(src, "node_modules")
      break
    }
  }
  return nodeModules
}

export const createStageRouteImportManifest: Stage & {overrideTriage: OverrideTriage} = ({
  getRouteCache,
  config,
}) => {
  const routeCache = getRouteCache()

  const routes: Record<string, Route> = {}

  const dotBlitz = join(findNodeModulesRoot(config.src), ".blitz")

  const writeManifestImplementation = makeDebouncedWriter(join(dotBlitz, "index.js"))
  const writeManifestBrowserImplementation = makeDebouncedWriter(join(dotBlitz, "index-browser.js"))
  const writeManifestDeclaration = makeDebouncedWriter(join(dotBlitz, "index.d.ts"))

  const stream = transform.file((file) => {
    if (!isPage(file.relative)) {
      return file
    }

    const defaultExportName = parseDefaultExportName(file.contents?.toString() ?? "")
    if (!defaultExportName) {
      return file
    }

    const entry = routeCache.get(file)
    if (!entry) {
      return file
    }

    function addRoute(path: string, defaultExportName: string) {
      routes[path] = {
        ...parseParametersFromRoute(path),
        name: defaultExportName,
      }

      const {declaration, implementation} = generateManifest(routes)

      writeManifestImplementation(implementation)
      writeManifestBrowserImplementation(implementation)
      writeManifestDeclaration(declaration)
    }

    addRoute(entry.uri, defaultExportName)

    return file
  })

  return {stream}
}

createStageRouteImportManifest.overrideTriage = (file: File) => {
  if (isPage(file.relative)) {
    return "proceed"
  }

  return undefined
}
